/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

/* The majority of this file came from Kris Kennaway <kris at fr.eebsd.org>. */

/* 0.4.28.c17 */
#include "sh_sys_types.h"

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "sh_sys_sockio.h"
#include <errno.h>
#include "sh_err.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lib.h"

#ifndef WIN32
int get_interfaces(int sockfd, struct ifconf *ifc)
{
  int      len, lastlen;
  char      *buf;

  /* Grab a list of interfaces */

  GD_S(2, "sockfd="); GD_I(2, sockfd); GD_S(2, " ifc="); GD_P(2, ifc);
  GD_S(2, "\n");
  
  lastlen = 0;
  len = 20 * sizeof(struct ifreq);  /* initial buffer size guess */
  for ( ; ; ) {
    buf = (char *)ymaloc(len, 325);

    ifc->ifc_len = len;
    ifc->ifc_buf = buf;
    if (ioctl(sockfd, SIOCGIFCONF, ifc) < 0) {
      if (errno != EINVAL || lastlen != 0) {
	GD_S(1, "ioctl(SIOCGIFCONF)\n");
	return -1;
      }
    } else {
      if (ifc->ifc_len == lastlen)
	break;    /* success, len has not changed */
      lastlen = ifc->ifc_len;
    }
    len += sizeof(struct ifreq);  /* increment */
    fre_str(&buf, 382);
  }
  
  GD_S(2, "get_interfaces returning success\n");
  return 0;
}
#endif

struct sockaddr_in *get_if_addr(char *iface)
{
#ifndef WIN32
  int family = AF_INET;
  int      sockfd, len, flags;
  char      *ptr;
  struct ifconf    ifc;
  struct ifreq    *ifr=0, ifrcopy;
  struct sockaddr_in  *sinptr=0;

  if (iface) {
    GD_S(2, "get_if_addr entering iface="); GD_S(2, iface); GD_S(2, "\n");
  }

  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    GD_S(1, "socket\n");
    return 0;
  }
  
  if (get_interfaces(sockfd, &ifc) == -1) {
    GD_S(1, "get_interfaces\n");
    return 0;
  }

  /* Walk the list of interfaces and locate the first one which is up */

  for (ptr = ifc.ifc_buf; ptr < (ifc.ifc_buf + ifc.ifc_len); ) {
    ifr = (struct ifreq *) ptr;

    GD_S(2, "ifr="); GD_P(2, ifr); GD_S(2, "  ifc.ifc_buf=");
    GD_P(2, ifc.ifc_buf); GD_S(2, "  ifc.ifc_len=");
    GD_I(2, ifc.ifc_len); GD_S(2, "\n");
    
    len = sizeof(struct ifreq);
    ptr += len;  /* for next one in buffer */

    if (ifr->ifr_addr.sa_family != family)
      continue;  /* ignore if not desired address family */

    ifrcopy = *ifr;
    if (ioctl(sockfd, SIOCGIFFLAGS, &ifrcopy) < 0) {
      GD_S(1, "ioctl(SIOCGIFFLAGS)\n");
    }

    flags = ifrcopy.ifr_flags;

    if ((flags & IFF_UP) == 0)
      continue;  /* ignore if interface not up */

    if (strncmp("lo", ifr->ifr_name, 2)==0)
      continue;

    if (iface) {
      if (!strncmp(iface, ifr->ifr_name, IFNAMSIZ) == 0)
	continue; /* Not the interface we're looking for */
    }

    switch (ifr->ifr_addr.sa_family) {
    case AF_INET:
      sinptr = (struct sockaddr_in *) &ifr->ifr_addr;
      printf("Local IP: %s\n", inet_ntoa(sinptr->sin_addr));
      break;

    default:
      printf("Protocol family not supported.");
    }
    break;
  }
  if (iface && (sinptr == 0)) {
    GD_S(1, "No such interface '"); GD_S(1, iface); GD_S(1, "'\n");
  }
  return sinptr;
#else
  return 0;
#endif
}
